from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from django.core.exceptions import ValidationError
from django.forms.fields import ChoiceField, TypedMultipleChoiceField

from .utils import fqn, stringify

if TYPE_CHECKING:
    from .registry import Registry


class StrategyFormField(ChoiceField):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.registry = kwargs.pop("registry")
        self.empty_value = kwargs.pop("empty_value", "")
        super().__init__(*args, **kwargs)

    def prepare_value(self, value: str | type) -> str | None:
        if isinstance(value, str):
            return value
        if value:
            return fqn(value)
        return None

    def valid_value(self, value: str) -> bool:
        return value in self.registry

    def _coerce(self, value: str) -> Any:
        if value == self.empty_value or value in self.empty_values:
            return self.empty_value
        try:
            v = self.to_python(value)
            if v in self.registry:
                return v
            raise ValidationError
        except (ValueError, TypeError, ValidationError):
            raise ValidationError(
                self.error_messages["invalid_choice"],
                code="invalid_choice",
                params={"value": f"'{value}'"},
            ) from None

    def clean(self, value: Any) -> type:
        value = super().clean(value)
        return self._coerce(value)


class StrategyMultipleChoiceFormField(TypedMultipleChoiceField):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.registry: Registry = kwargs.pop("registry")
        kwargs["coerce"] = self.coerce
        super().__init__(*args, **kwargs)

    def prepare_value(self, value: str | Sequence[str]) -> list[str] | None:
        ret = value
        if isinstance(value, (list, tuple)):
            ret = stringify(value)
        if ret:
            return ret.split(",")

    def coerce(self, value: str) -> type | None:
        return self.registry.get_by_name(value)

    def valid_value(self, value: str) -> bool:
        return value in self.registry



from __future__ import annotations

import logging
from inspect import isclass
from typing import TYPE_CHECKING, Any, Protocol, Sequence

from django import forms
from django.core.exceptions import ValidationError
from django.db.models import Field, Model
from django.db.models.fields import BLANK_CHOICE_DASH
from django.db.models.lookups import (
    Contains,
    Exact,
    IContains,
    IExact,
    Lookup,
)
from django.utils.text import capfirst

from .exceptions import StrategyClassError, StrategyNameError
from .forms import StrategyFormField, StrategyMultipleChoiceFormField
from .utils import fqn, get_class, stringify
from .validators import ClassnameValidator, RegistryValidator

if TYPE_CHECKING:
    from django.db.backends.base.base import BaseDatabaseWrapper
    from django.db.models.base import ModelBase
    from django.db.models.fields import _ChoicesList, _LimitChoicesTo
    from django.utils.choices import BlankChoiceIterator

NOCONTEXT = object()

logger = logging.getLogger(__name__)


class StrategyDescriptor(Protocol):
    def __init__(self, field: AbstractStrategyField) -> None: ...  # pragma: no cover
    def __get__(self, instance: Any, owner: type[BaseDatabaseWrapper]) -> Any: ...  # pragma: no cover
    def __set__(self, obj: Model, original: str) -> None: ...  # pragma: no cover


class StrategyClassFieldDescriptor:
    def __init__(self, field: StrategyField) -> None:
        self.field = field

    def __get__(self, obj: Model | None, value: type[Model] | None = None) -> StrategyClassField | None:
        if obj is None:
            return None
        return obj.__dict__.get(self.field.name)

    def __set__(self, obj: Model, original: str) -> None:
        if not original:
            value = None
        else:
            try:
                value = get_class(original)
            except (
                AttributeError,
                ModuleNotFoundError,
                ImportError,
                StrategyNameError,
            ) as e:
                if callable(self.field.import_error):
                    value = self.field.import_error(original, e)
                else:
                    value = self.field.import_error
            except Exception as e:  # pragma: no-cover
                logger.exception(e)
                raise ValidationError(original) from e

        obj.__dict__[self.field.name] = value
        try:
            raw_value = fqn(original or "")
        except StrategyClassError:
            raw_value = None
        obj.__dict__[f"_strategy_fqn_{self.field.name}"] = raw_value


class MultipleStrategyClassFieldDescriptor:
    def __init__(self, field: StrategyField) -> None:
        self.field = field

    def __get__(self, obj: Model, __: type[ModelBase] | None = None) -> list[type] | None:
        if obj is None:
            return None
        value = obj.__dict__.get(self.field.name)
        if value is None:
            return None
        if isinstance(value, str):
            value = value.split(",")
        if not isinstance(value, (list, tuple)):
            value = [value] if value is not None else None
        ret = []
        for v in value:
            if v:
                try:
                    v1 = get_class(v)
                    ret.append(v1)
                except StrategyNameError as e:
                    if callable(self.field.import_error):
                        return self.field.import_error(value, e)
                    return self.field.import_error

        return ret

    def __set__(self, obj: Model, value: Any) -> None:
        obj.__dict__[self.field.name] = value


class AbstractStrategyField(Field):
    descriptor: type[StrategyDescriptor]

    registry = None
    form_class: type[StrategyFormField | StrategyMultipleChoiceFormField]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.import_error = kwargs.pop("import_error", None)
        kwargs["max_length"] = 200

        self.registry = kwargs.pop("registry", None)
        super().__init__(*args, **kwargs)
        self.validators.append(ClassnameValidator(None))
        if self.registry:
            self.validators.append(RegistryValidator(self.registry))

    @property
    def flatchoices(self) -> None:  # type:ignore[override] #  this is needed to display in admin #
        return None

    def contribute_to_class(self, cls: type[Model], name: str, private_only: bool = False) -> None:
        self.set_attributes_from_name(name)
        self.model = cls
        if callable(self.registry):
            self.registry = self.registry(cls)
        cls._meta.add_field(self)
        setattr(cls, self.name, self.descriptor(self))

    def deconstruct(self) -> tuple[str, str, Sequence[Any], dict[str, Any]]:
        name, path, args, kwargs = super().deconstruct()
        del kwargs["max_length"]
        if "registry" in kwargs:
            del kwargs["registry"]
        if "choices" in kwargs:
            del kwargs["choices"]
        return name, path, args, kwargs

    def get_db_prep_value(self, value: Any, connection: BaseDatabaseWrapper, prepared: bool = False) -> Any:
        return super().get_db_prep_value(value, connection, prepared)

    def get_db_prep_save(self, value: Any, connection: BaseDatabaseWrapper) -> Any:
        return super().get_db_prep_value(value, connection)

    def get_prep_value(self, value: Any) -> Any:
        if value is None:
            return None
        return fqn(value)

    def value_to_string(self, obj: Model) -> str:
        value = self.value_from_object(obj)
        return fqn(value)

    def get_internal_type(self) -> str:
        return "CharField"

    def _check_choices(self) -> list:
        return []

    def _get_choices(self) -> list[tuple[AbstractStrategyField, str]]:
        if self.registry:
            return self.registry.as_choices()
        return []

    def _set_choices(self, value: tuple) -> None:
        pass

    choices = property(_get_choices, _set_choices)

    def get_choices(
        self,
        include_blank: bool = True,
        blank_choice: _ChoicesList = BLANK_CHOICE_DASH,
        limit_choices_to: _LimitChoicesTo | None = None,
        ordering: Sequence[str] = (),
    ) -> BlankChoiceIterator | _ChoicesList:
        first_choice = blank_choice if include_blank else []

        return first_choice + self.choices

    def validate(self, value: Any, model_instance: Model | None) -> None:
        if fqn(value) not in self.registry:
            raise ValidationError(f"{value} is not a valid choice")

    def formfield(
        self,
        form_class: type[forms.Field] | None = None,
        choices_form_class: type[forms.ChoiceField] | None = None,
        **kwargs: Any,
    ) -> forms.Field | None:
        defaults = {
            "required": not self.blank,
            "label": capfirst(self.verbose_name),
            "help_text": self.help_text,
            "registry": self.registry,
        }
        if self.has_default():
            if callable(self.default):
                defaults["initial"] = self.default
                defaults["show_hidden_initial"] = True
            else:
                defaults["initial"] = self.get_default()
        include_blank = self.blank or not (self.has_default() or "initial" in kwargs)
        defaults["choices"] = self.get_choices(include_blank=include_blank)
        if self.null:
            defaults["empty_value"] = None
        form_class = choices_form_class or self.form_class
        for k in list(kwargs):
            if k not in (
                "empty_value",
                "required",
                "choices",
                "registry",
                "widget",
                "label",
                "initial",
                "help_text",
                "error_messages",
                "show_hidden_initial",
            ):
                del kwargs[k]
        defaults.update(kwargs)
        return form_class(**defaults)


class RegexFormField(forms.CharField):
    pass


class StrategyClassField(AbstractStrategyField):
    form_class = StrategyFormField
    descriptor = StrategyClassFieldDescriptor

    def validate(self, value: Any, model_instance: Model | None) -> None:
        if fqn(value) not in self.registry:
            raise ValidationError(f"{value} is not a valid choice")


class MultipleStrategyClassField(AbstractStrategyField):
    descriptor = MultipleStrategyClassFieldDescriptor
    form_class = StrategyMultipleChoiceFormField

    def validate(self, values: Any, model_instance: Model | None) -> None:
        for value in values:
            if value not in self.registry:
                raise ValidationError(f"{value} is not a valid choice")

    def get_db_prep_save(self, value: Any, connection: BaseDatabaseWrapper, prepared: bool = False) -> Any:
        value = list(filter(lambda x: x, value)) if value is not None else None
        return super().get_db_prep_save(value, connection)

    def get_prep_value(self, value: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, (list, tuple)):
            return stringify(value)
        if isinstance(value, str):
            return value
        return None

    def get_lookup(self, lookup_name: str) -> type[Lookup] | None:
        if lookup_name == "in":
            raise TypeError(f"Lookup type {lookup_name} not supported.")
        return super().get_lookup(lookup_name)

    def get_choices(
        self,
        include_blank: bool = True,
        blank_choice: _ChoicesList = BLANK_CHOICE_DASH,
        limit_choices_to: _LimitChoicesTo | None = None,
        ordering: Sequence[str] = (),
    ) -> BlankChoiceIterator | _ChoicesList:
        return AbstractStrategyField.get_choices(self, False, blank_choice)


class StrategyFieldDescriptor(StrategyClassFieldDescriptor):
    def __get__(self, obj: Model | None, value: type[Model] | None = None) -> StrategyField | None:
        if obj is None:
            return None
        return obj.__dict__.get(self.field.name)

    def __set__(self, obj: Model, original: str | None) -> None:
        if not original:
            value = None
        else:
            try:
                value = get_class(original)
            except (
                AttributeError,
                ModuleNotFoundError,
                ImportError,
                StrategyNameError,
            ) as e:
                if callable(self.field.import_error):
                    value = self.field.import_error(original, e)
                else:
                    value = self.field.import_error
            except Exception as e:  # pragma: no-cover
                logger.exception(e)
                raise ValidationError(original) from e

        if isclass(value):
            value = self.field.factory(value, obj)

        obj.__dict__[self.field.name] = value
        try:
            raw_value = fqn(original or "")
        except StrategyClassError:
            raw_value = None
        obj.__dict__[f"_strategy_fqn_{self.field.name}"] = raw_value


class StrategyField(StrategyClassField):
    descriptor = StrategyFieldDescriptor

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.factory = kwargs.pop("factory", lambda klass, obj: klass(obj))
        super().__init__(*args, **kwargs)

    def validate(self, value: Any, model_instance: Model | None) -> None:
        if fqn(value) not in self.registry:
            raise ValidationError(f"{value} is not a valid choice")

    def pre_save(self, model_instance: Model, add: bool) -> str | None:
        value = getattr(model_instance, self.attname)
        if value:
            return fqn(value)
        return None


class MultipleStrategyFieldDescriptor(MultipleStrategyClassFieldDescriptor):
    def __get__(self, obj: Model, __: type[ModelBase] | None = None) -> list[type]:
        if obj is None:
            return []
        value = obj.__dict__.get(self.field.name)

        if value and isinstance(value, (list, tuple, str)):
            ret = []
            if isinstance(value, str):
                value = value.split(",")
            for v in value:
                try:
                    cleaned = get_class(v)
                    ret.append(self.field.factory(cleaned, obj))
                except (  # noqa: PERF203
                    AttributeError,
                    ModuleNotFoundError,
                    ImportError,
                    StrategyNameError,
                ) as e:
                    if callable(self.field.import_error):
                        value = self.field.import_error(value, e)
                    else:
                        value = self.field.import_error
                except Exception as e:  # pragma: no-cover
                    logger.exception(e)
                    raise ValidationError(value) from e
            return ret
        return []

    def __set__(self, obj: Model, value: Any) -> None:
        obj.__dict__[self.field.name] = value


class MultipleStrategyField(MultipleStrategyClassField):
    descriptor = MultipleStrategyFieldDescriptor

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.factory = kwargs.pop("factory", lambda klass, obj: klass(obj))
        super().__init__(*args, **kwargs)

    def validate(self, values: Any, model_instance: Model | None) -> None:
        for value in values:
            if type(value) not in self.registry:
                raise ValidationError(f"{value} is not a valid choice")

    def get_lookup(self, lookup_name: str) -> type[Lookup[Any]] | None:
        if lookup_name == "in":
            raise TypeError(f"Lookup type {lookup_name} not supported.")
        return super().get_lookup(lookup_name)


class StrategyFieldLookupMixin(Lookup):
    def get_prep_lookup(self) -> str | None:
        value = super().get_prep_lookup()
        if value is None:
            return None
        if isinstance(value, str):
            pass
        elif isinstance(value, (list, tuple)):
            value = stringify(value)
        elif isclass(value) or isinstance(value, (self.lhs.output_field.registry.klass, object)):
            value = fqn(value)
        return value


class MultipleStrategyFieldLookupMixin(Lookup):
    def get_prep_lookup(self) -> str | None:
        value = super().get_prep_lookup()
        if value is None:
            return None
        if isinstance(value, str):
            pass
        elif isinstance(value, (list, tuple)):
            value = stringify(value)
        elif isclass(value) or isinstance(value, (self.lhs.output_field.registry.klass, object)):
            value = fqn(value)
        return value


class StrategyFieldContains(StrategyFieldLookupMixin, IContains):
    pass


class StrategyFieldIContains(StrategyFieldLookupMixin, Contains):
    pass


class MultipleStrategyFieldContains(MultipleStrategyFieldLookupMixin, Contains):
    pass


class MultipleStrategyFieldExact(MultipleStrategyFieldLookupMixin, Exact):
    pass


class MultipleStrategyFieldIExact(MultipleStrategyFieldLookupMixin, IExact):
    pass


StrategyField.register_lookup(StrategyFieldContains)
StrategyField.register_lookup(StrategyFieldIContains)

StrategyClassField.register_lookup(StrategyFieldContains)
StrategyClassField.register_lookup(StrategyFieldIContains)

MultipleStrategyClassField.register_lookup(MultipleStrategyFieldContains)
MultipleStrategyClassField.register_lookup(MultipleStrategyFieldExact)
MultipleStrategyClassField.register_lookup(MultipleStrategyFieldIExact)
MultipleStrategyField.register_lookup(MultipleStrategyFieldContains)



from __future__ import annotations

import importlib
import logging
import types
from inspect import isclass
from typing import Any, Sequence

from django.utils.module_loading import import_string

from . import config
from .exceptions import StrategyAttributeError, StrategyClassError, StrategyNameError

logger = logging.getLogger(__name__)


class ModulesCache(dict):
    def __missing__(self, name: str) -> Any:
        if "." not in name:
            raise StrategyNameError(name)

        module_path, class_str = name.rsplit(".", 1)
        module = importlib.import_module(module_path)
        try:
            handler = getattr(module, class_str)
            self[name] = handler
            return handler
        except AttributeError:
            raise StrategyAttributeError(name, module_path, class_str) from None


_cache = ModulesCache()


def default_classloader(value: str | type | None) -> Any:
    if not value:
        return None
    if isinstance(value, str):
        return import_by_name(value)
    if isclass(value):
        return value

    t = type(value)
    if t.__module__ in ("builtins", "__builtin__"):
        return None
    return t


importer = None


def get_class(value: str) -> Any:
    global importer  # noqa: PLW0603
    if importer is None:
        importer = import_string(config.CLASSLOADER)
    return importer(value)


def get_display_string(klass: type, display_attribute: str | None = None) -> str:
    if display_attribute and hasattr(klass, display_attribute):
        attr = getattr(klass, display_attribute)
        if attr is None:
            return fqn(klass)
        if callable(attr):
            return str(attr())
        return str(attr)

    return fqn(klass)


def get_attr(obj: Any, attr: str, default: Any = None) -> Any:
    """Recursive get object's attribute. May use dot notation."""
    if "." not in attr:
        return getattr(obj, attr, default)
    parts = attr.split(".")
    return get_attr(getattr(obj, parts[0], default), ".".join(parts[1:]), default)


def fqn(o: Any) -> str:
    """Returns the fully qualified class name of an object or a class

    :param o: object or class
    :return: class name
    """
    parts = []
    if isinstance(o, str):
        return o
    if not hasattr(o, "__module__"):
        raise StrategyClassError(o)
    parts.append(o.__module__)
    if isclass(o) or isinstance(o, types.FunctionType):
        parts.append(o.__name__)
    else:
        parts.append(o.__class__.__name__)
    return ".".join(parts)


def import_by_name(name: str) -> Any:
    """dynamically load a class from a string

    es:
        klass = import_by_name('my_package.my_module.my_class')
        some_object = klass()

    :param name:
    :return:

    """
    return _cache[name]


def stringify(value: Sequence[Any]) -> str:
    ret = []
    for v in value:
        if isinstance(v, str) and v:
            ret.append(v)
        else:
            ret.append(fqn(v))
    return ",".join(sorted(ret))



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ClassVar, Sequence

from django.core.validators import BaseValidator
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from ..fields import AbstractStrategyField, ClassnameValidator
from ..utils import fqn, import_by_name

if TYPE_CHECKING:
    from ..registry import Registry


logger = logging.getLogger(__name__)


class RegistryValidator(BaseValidator):
    def __call__(self, value: Sequence[str]) -> None:
        if not isinstance(value, (list, tuple)):
            value = [value]
        for entry in value:
            if not self.limit_value.is_valid(entry):
                raise ValidationError(f"Invalid entry `{fqn(entry)}`")


class DrfStrategyField(serializers.ChoiceField):
    default_validators: ClassVar[list] = [ClassnameValidator]

    def __init__(self, registry: Registry, **kwargs: Any) -> None:
        choices = registry.as_choices()
        super().__init__(choices, **kwargs)
        self.registry = registry

    def get_validators(self) -> Sequence[BaseValidator]:
        ret = super().get_validators()
        ret.append(RegistryValidator(self.registry))
        return ret

    def to_representation(self, obj: AbstractStrategyField) -> str:
        return fqn(obj)

    def to_internal_value(self, data: Any) -> Any:
        return data


class DrfMultipleStrategyField(serializers.MultipleChoiceField):
    default_validators: ClassVar[list] = [ClassnameValidator]

    def __init__(self, registry: Registry, **kwargs: Any) -> None:
        choices = registry.as_choices()
        self.registry = registry
        super().__init__(choices=choices, **kwargs)

    def get_validators(self) -> list[BaseValidator]:
        ret = super().get_validators()
        ret.append(RegistryValidator(self.registry))
        return ret

    def to_representation(self, obj: list[type]) -> list[str]:
        return [fqn(i) for i in obj]

    def to_internal_value(self, data: list[str]) -> list[type]:
        return [import_by_name(i) for i in data]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django.core.exceptions import ValidationError
from django.core.validators import BaseValidator
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _

from .exceptions import StrategyNameError
from .utils import get_class

if TYPE_CHECKING:
    from .registry import Registry


@deconstructible
class ClassnameValidator(BaseValidator):
    message = _("Ensure this value is valid class name (it is %(show_value)s).")
    code = "classname"

    def __call__(self, value: Any) -> None:
        cleaned = self.clean(value)
        params = {"show_value": cleaned}
        try:
            get_class(cleaned)
        except (ImportError, TypeError, StrategyNameError):
            raise ValidationError(self.message, code=self.code, params=params) from None


@deconstructible
class RegistryValidator(ClassnameValidator):
    message = _("Invalid entry `%(show_value)s`")
    message_many = _("Invalid entries `%(show_value)s`")

    code = "registry"

    def __init__(self, registry: Registry, message: str | None = None) -> None:
        super().__init__(registry, message)
        self.registry = registry  # aliasing self.limit_value for readability

    def __call__(self, value: Any) -> None:
        cleaned = self.clean(value)
        params = {"show_value": cleaned}
        try:
            if isinstance(value, (list, tuple)):
                errs = [c for c in cleaned if c not in self.registry]
                if len(errs) == 1:
                    raise ValidationError(self.message, code=self.code, params={"show_value": errs[0]}) from None
                if len(errs) > 1:
                    raise ValidationError(
                        self.message_many, code=self.code, params={"show_value": ", ".join(errs)}
                    ) from None
            elif not issubclass(get_class(value), self.registry.klass):
                raise ValidationError(self.message, code=self.code, params=params) from None

        except (ImportError, TypeError, StrategyNameError):
            raise ValidationError(self.message, code=self.code, params=params) from None



from __future__ import annotations


class StrategyNameError(ValueError):
    default_message = "Invalid value '%s': must be a valid python dotted name."

    def __init__(self, name: str, message: str | None = None) -> None:
        self.name = str(name)
        self.message = message or self.default_message

    def __repr__(self) -> str:
        return self.message % self.name


class StrategyClassError(ValueError):
    default_message = "Invalid argument: '%s' is a invalid python name"

    def __init__(self, name: str, message: str | None = None) -> None:
        self.name = str(name)
        self.message = message or self.default_message

    def __repr__(self) -> str:
        return self.message % self.name


class StrategyImportError(ImportError):
    pass


class StrategyAttributeError(AttributeError):
    default_message = "Unable to import %(name)s. %(module)s does not have %(class_str)s attribute"

    def __init__(self, name: str, module_path: str, class_str: str, message: str | None = None) -> None:
        self.name = str(name)
        self.module_path = module_path
        self.class_str = class_str
        self.message = message or self.default_message

    def __repr__(self) -> str:
        return self.message % (self.name, self.module_path, self.class_str)



from __future__ import annotations

import logging
from inspect import isclass
from typing import Any

from django.utils.functional import cached_property

from .utils import fqn, get_class, get_display_string, import_by_name

logger = logging.getLogger(__name__)


class Registry(list):
    def __init__(self, base_class: type, *args: Any, **kwargs: Any) -> None:
        self._klass = base_class
        self._label_attribute = kwargs.get("label_attribute")
        self._choices = None
        list.__init__(self, *args[:])

    @cached_property
    def klass(self) -> Any:
        if isinstance(self._klass, str):
            return import_by_name(self._klass)
        return self._klass

    def get_name(self, entry: type) -> str:
        return get_display_string(entry, self._label_attribute)

    def get_by_name(self, entry: str) -> Any:
        return get_class(entry)

    def is_valid(self, value: str) -> bool:
        if value and isinstance(value, str):
            try:
                value = import_by_name(value)
            except (ImportError, ValueError, AttributeError):
                return False

        if self.klass:
            return (isclass(value) and issubclass(value, self.klass)) or (isinstance(value, self.klass))

        return True

    def as_choices(self) -> tuple[str, str]:
        if not self._choices:
            self._choices = sorted((fqn(klass), self.get_name(klass)) for klass in self)
        return self._choices

    def register(self, class_or_fqn: type | str) -> Any:
        cls = import_by_name(class_or_fqn) if isinstance(class_or_fqn, str) else class_or_fqn

        if cls == self.klass:
            return None

        if self.klass and not issubclass(cls, self.klass):
            raise ValueError(f"'{class_or_fqn}' is not a subtype of {self.klass}")

        if cls in self:
            return None

        super().append(cls)
        self._choices = None
        return class_or_fqn

    def append(self, class_or_fqn: type | str) -> None:
        self.register(class_or_fqn)

    def __contains__(self, y: Any) -> bool:
        if isinstance(y, str):
            try:
                return super().__contains__(import_by_name(y))
            except (ImportError, ValueError):
                return False
        elif isclass(y):
            return super().__contains__(y)
        elif (self._klass and isinstance(y, self._klass)) or isinstance(y, object):
            return super().__contains__(type(y))
        return super().__contains__(y)
